import pytest
from testing_lib import filecases
from wheel_inspect.classes import WheelFile
from wheel_inspect.errors import WheelValidationError
from wheel_inspect.inspecting import verify_record


@pytest.mark.parametrize("whlfile,expected", filecases("bad-wheels", "*.whl"))
def test_verify_bad_wheels(whlfile, expected):
    with WheelFile(whlfile) as whl:
        with pytest.raises(WheelValidationError) as excinfo:
            verify_record(whl, whl.get_record())
        assert type(excinfo.value).__name__ == expected["type"]
        assert str(excinfo.value) == expected["str"]



import json
from operator import attrgetter
from pathlib import Path
from jsonschema import validate
import pytest
from testing_lib import filecases
from wheel_inspect import (
    DIST_INFO_SCHEMA,
    WHEEL_SCHEMA,
    inspect_dist_info_dir,
    inspect_wheel,
)


@pytest.mark.parametrize("whlfile,expected", filecases("wheels", "*.whl"))
def test_inspect_wheel(whlfile, expected):
    inspection = inspect_wheel(whlfile)
    assert inspection == expected
    validate(inspection, WHEEL_SCHEMA)


@pytest.mark.parametrize(
    "didir",
    [
        p
        for p in (Path(__file__).with_name("data") / "dist-infos").iterdir()
        if p.is_dir()
    ],
    ids=attrgetter("name"),
)
def test_inspect_dist_info_dir(didir):
    with open(str(didir) + ".json") as fp:
        expected = json.load(fp)
    inspection = inspect_dist_info_dir(didir)
    assert inspection == expected
    validate(inspection, DIST_INFO_SCHEMA)



import io
import entry_points_txt
from readme_renderer.rst import render
from . import errors
from .classes import DistInfoDir, FileProvider, WheelFile
from .util import (
    extract_modules,
    is_dist_info_path,
    split_content_type,
    split_keywords,
    unique_projects,
    yield_lines,
)


def parse_entry_points(fp):
    """
    Parse the contents of a text filehandle ``fp`` as an
    :file:`entry_points.txt` file and return a `dict` that maps entry point
    group names to sub-`dict`s that map entry point names to sub-sub-`dict`s
    with ``"module"``, ``"attr"``, and ``"extras"`` keys.

    For example, the following input:

    .. code-block:: ini

        [console_scripts]
        do-thing = pkg.main:__main__

        [plugin.point]
        plug-thing = pkg.plug [xtra]

    would be parsed into the following structure::

        {
            "console_scripts": {
                "do-thing": {
                    "module": "pkg.main",
                    "attr": "__main__",
                    "extras": []
                }
            },
            "plugin.point": {
                "plug-thing": {
                    "module": "pkg.plug",
                    "attr": None,
                    "extras": ["xtra"]
                }
            }
        }
    """
    epset = entry_points_txt.load(fp)
    return {
        gr: {
            k: {
                "module": e.module,
                "attr": e.attr,
                "extras": list(e.extras),
            }
            for k, e in eps.items()
        }
        for gr, eps in epset.items()
    }


def readlines(fp):
    return list(yield_lines(fp))


EXTRA_DIST_INFO_FILES = [
    # file name, handler function, result dict key
    # <https://setuptools.readthedocs.io/en/latest/formats.html>:
    ("dependency_links.txt", readlines, "dependency_links"),
    ("entry_points.txt", parse_entry_points, "entry_points"),
    ("namespace_packages.txt", readlines, "namespace_packages"),
    ("top_level.txt", readlines, "top_level"),
]


def inspect(obj):  # (DistInfoProvider) -> dict
    about = obj.basic_metadata()
    about["dist_info"] = {}
    about["valid"] = True
    has_dist_info = True

    try:
        record = obj.get_record()
    except errors.WheelValidationError as e:
        about["valid"] = False
        about["validation_error"] = {
            "type": type(e).__name__,
            "str": str(e),
        }
        has_dist_info = not isinstance(e, errors.DistInfoError)
    else:
        about["dist_info"]["record"] = record.for_json()
        if isinstance(obj, FileProvider):
            try:
                verify_record(obj, record)
            except errors.WheelValidationError as e:
                about["valid"] = False
                about["validation_error"] = {
                    "type": type(e).__name__,
                    "str": str(e),
                }

    if has_dist_info:
        try:
            metadata = obj.get_metadata()
        except errors.WheelValidationError as e:
            metadata = {}
            about["valid"] = False
            about["validation_error"] = {
                "type": type(e).__name__,
                "str": str(e),
            }
        else:
            about["dist_info"]["metadata"] = metadata

        try:
            about["dist_info"]["wheel"] = obj.get_wheel_info()
        except errors.WheelValidationError as e:
            about["valid"] = False
            about["validation_error"] = {
                "type": type(e).__name__,
                "str": str(e),
            }

        for fname, parser, key in EXTRA_DIST_INFO_FILES:
            try:
                with obj.open_dist_info_file(fname) as binfp, io.TextIOWrapper(
                    binfp, "utf-8"
                ) as txtfp:
                    about["dist_info"][key] = parser(txtfp)
            except errors.MissingDistInfoFileError:
                pass

        if obj.has_dist_info_file("zip-safe"):
            about["dist_info"]["zip_safe"] = True
        elif obj.has_dist_info_file("not-zip-safe"):
            about["dist_info"]["zip_safe"] = False

    else:
        metadata = {}

    about["derived"] = {
        "description_in_body": "BODY" in metadata,
        "description_in_headers": "description" in metadata,
    }

    if "BODY" in metadata and "description" not in metadata:
        metadata["description"] = metadata["BODY"]
    metadata.pop("BODY", None)
    readme = metadata.get("description")
    if readme is not None:
        metadata["description"] = {"length": len(metadata["description"])}
        dct = metadata.get("description_content_type")
        if dct is None or split_content_type(dct)[:2] == ("text", "x-rst"):
            about["derived"]["readme_renders"] = render(readme) is not None
        else:
            about["derived"]["readme_renders"] = True
    else:
        about["derived"]["readme_renders"] = None

    if metadata.get("keywords") is not None:
        (
            about["derived"]["keywords"],
            about["derived"]["keyword_separator"],
        ) = split_keywords(metadata["keywords"])
    else:
        about["derived"]["keywords"], about["derived"]["keyword_separator"] = [], None
    about["derived"]["keywords"] = sorted(set(about["derived"]["keywords"]))

    about["derived"]["dependencies"] = sorted(
        unique_projects(req["name"] for req in metadata.get("requires_dist", []))
    )

    about["derived"]["modules"] = extract_modules(
        [rec["path"] for rec in about["dist_info"].get("record", [])]
    )

    return about


def inspect_wheel(path):
    """
    Examine the Python wheel at the given path and return various information
    about the contents within as a JSON-serializable `dict`
    """
    with WheelFile(path) as wf:
        return inspect(wf)


def inspect_dist_info_dir(path):
    """
    Examine the ``*.dist-info`` directory at the given path and return various
    information about the contents within as a JSON-serializable `dict`
    """
    with DistInfoDir(path) as did:
        return inspect(did)


def verify_record(fileprod: FileProvider, record):
    files = set(fileprod.list_files())
    # Check everything in RECORD against actual values:
    for entry in record:
        if entry.path.endswith("/"):
            if not fileprod.has_directory(entry.path):
                raise errors.FileMissingError(entry.path)
        elif entry.path not in files:
            raise errors.FileMissingError(entry.path)
        elif entry.digest is not None:
            file_size = fileprod.get_file_size(entry.path)
            if entry.size != file_size:
                raise errors.RecordSizeMismatchError(
                    entry.path,
                    entry.size,
                    file_size,
                )
            digest = fileprod.get_file_hash(entry.path, entry.digest_algorithm)
            if digest != entry.digest:
                raise errors.RecordDigestMismatchError(
                    entry.path,
                    entry.digest_algorithm,
                    entry.digest,
                    digest,
                )
        elif not is_dist_info_path(entry.path, "RECORD"):
            raise errors.NullEntryError(entry.path)
        files.discard(entry.path)
    # Check that the only files that aren't in RECORD are signatures:
    for path in files:
        if not is_dist_info_path(entry.path, "RECORD.jws") and not is_dist_info_path(
            entry.path, "RECORD.p7s"
        ):
            raise errors.ExtraFileError(path)



import base64
from binascii import hexlify, unhexlify
from collections import OrderedDict
import csv
import hashlib
import re
import attr
from . import errors


@attr.s
class Record:
    files = attr.ib()

    def __iter__(self):
        return iter(self.files.values())

    def __contains__(self, filename):
        return filename in self.files

    def for_json(self):
        return [e.for_json() for e in self.files.values()]


@attr.s
class RecordEntry:
    path = attr.ib()
    digest_algorithm = attr.ib()
    #: The digest in hex format
    digest = attr.ib()
    size = attr.ib()

    @classmethod
    def from_csv_fields(cls, fields):
        try:
            path, alg_digest, size = fields
        except ValueError:
            raise errors.RecordLengthError(
                fields[0] if fields else None,
                len(fields),
            )
        if not path:
            raise errors.EmptyPathError()
        elif "//" in path or "." in path.split("/") or ".." in path.split("/"):
            raise errors.NonNormalizedPathError(path)
        elif path.startswith("/"):
            raise errors.AbsolutePathError(path)
        if alg_digest:
            digest_algorithm, digest = alg_digest.split("=", 1)
            if digest_algorithm not in hashlib.algorithms_guaranteed:
                raise errors.UnknownDigestError(path, digest_algorithm)
            elif digest_algorithm in ("md5", "sha1"):
                raise errors.WeakDigestError(path, digest_algorithm)
            sz = (getattr(hashlib, digest_algorithm)().digest_size * 8 + 5) // 6
            if not re.fullmatch(r"[-_0-9A-Za-z]{%d}" % (sz,), digest):
                raise errors.MalformedDigestError(path, digest_algorithm, digest)
            digest = record_digest2hex(digest)
        else:
            digest_algorithm, digest = None, None
        if size:
            try:
                size = int(size)
            except ValueError:
                raise errors.MalformedSizeError(path, size)
        else:
            size = None
        if digest is None and size is not None:
            raise errors.EmptyDigestError(path)
        elif digest is not None and size is None:
            raise errors.EmptySizeError(path)
        return cls(
            path=path,
            digest_algorithm=digest_algorithm,
            digest=digest,
            size=size,
        )

    def for_json(self):
        return {
            "path": self.path,
            "digests": (
                {self.digest_algorithm: hex2record_digest(self.digest)}
                if self.digest is not None
                else {}
            ),
            "size": self.size,
        }


def parse_record(fp):
    # Format defined in PEP 376
    files = OrderedDict()
    for fields in csv.reader(fp, delimiter=",", quotechar='"'):
        if not fields:
            continue
        entry = RecordEntry.from_csv_fields(fields)
        if entry.path in files and files[entry.path] != entry:
            raise errors.RecordConflictError(entry.path)
        files[entry.path] = entry
    return Record(files)


def hex2record_digest(data):
    return base64.urlsafe_b64encode(unhexlify(data)).decode("us-ascii").rstrip("=")


def record_digest2hex(data):
    pad = "=" * (4 - (len(data) & 3))
    return hexlify(base64.urlsafe_b64decode(data + pad)).decode("us-ascii")



import abc
import io
from pathlib import Path
from zipfile import ZipFile
from wheel_filename import WheelFilename
from . import errors
from .metadata import parse_metadata
from .record import parse_record
from .util import digest_file, find_dist_info_dir
from .wheel_info import parse_wheel_info


class DistInfoProvider(abc.ABC):
    """
    An interface for resources that are or contain a :file:`*.dist-info`
    directory
    """

    @abc.abstractmethod
    def basic_metadata(self):
        """
        Returns a `dict` of class-specific simple metadata about the resource
        """
        ...

    @abc.abstractmethod
    def open_dist_info_file(self, path):
        """
        Returns a readable binary IO handle for reading the contents of the
        file at the given path beneath the :file:`*.dist-info` directory
        """
        ### TODO: Specify here that MissingDistInfoFileError is raised if file
        ### not found?
        ...

    @abc.abstractmethod
    def has_dist_info_file(self, path):
        """
        Returns true iff a file exists at the given path beneath the
        :file:`*.dist-info` directory
        """
        ...

    def get_metadata(self):
        try:
            with (
                self.open_dist_info_file("METADATA") as binfp,
                io.TextIOWrapper(binfp, "utf-8") as txtfp,
            ):
                return parse_metadata(txtfp)
        except errors.MissingDistInfoFileError:
            raise errors.MissingMetadataError()

    def get_record(self):
        try:
            with (
                self.open_dist_info_file("RECORD") as binfp,
                io.TextIOWrapper(binfp, "utf-8", newline="") as txtfp,
            ):
                # The csv module requires this file to be opened with
                # `newline=''`
                return parse_record(txtfp)
        except errors.MissingDistInfoFileError:
            raise errors.MissingRecordError()

    def get_wheel_info(self):
        try:
            with (
                self.open_dist_info_file("WHEEL") as binfp,
                io.TextIOWrapper(binfp, "utf-8") as txtfp,
            ):
                return parse_wheel_info(txtfp)
        except errors.MissingDistInfoFileError:
            raise errors.MissingWheelInfoError()


class FileProvider(abc.ABC):
    @abc.abstractmethod
    def list_files(self):
        """
        Returns a list of files in the resource.  Each file is represented as a
        relative ``/``-separated path as would appear in a :file:`RECORD` file.
        Directories are not included in the list.

        :rtype: List[str]
        """
        ...

    @abc.abstractmethod
    def has_directory(self, path):
        """
        Returns true iff the directory at ``path`` exists in the resource.

        :param str path: a relative ``/``-separated path that ends with a ``/``
        :rtype: bool
        """
        ...

    @abc.abstractmethod
    def get_file_size(self, path):
        """
        Returns the size of the file at ``path`` in bytes.

        :param str path: a relative ``/``-separated path
        :rtype: int
        """
        ...

    @abc.abstractmethod
    def get_file_hash(self, path, algorithm):
        """
        Returns a hexdigest of the contents of the file at ``path`` computed
        using the digest algorithm ``algorithm``.

        :param str path: a relative ``/``-separated path
        :param str algorithm: the name of the digest algorithm to use, as
            recognized by `hashlib`
        :rtype: str
        """
        ...


class DistInfoDir(DistInfoProvider):
    def __init__(self, path):
        self.path = Path(path)

    def __enter__(self):
        return self

    def __exit__(self, _exc_type, _exc_value, _traceback):
        return False

    def basic_metadata(self):
        return {}

    def open_dist_info_file(self, path):
        # returns a binary IO handle; raises MissingDistInfoFileError if file
        # does not exist
        try:
            return (self.path / path).open("rb")
        except FileNotFoundError:
            raise errors.MissingDistInfoFileError(path)

    def has_dist_info_file(self, path):
        return (self.path / path).exists()


class WheelFile(DistInfoProvider, FileProvider):
    def __init__(self, path):
        self.path = Path(path)
        self.parsed_filename = WheelFilename.parse(self.path)
        self.fp = None
        self.zipfile = None
        self._dist_info = None

    def __enter__(self):
        self.fp = self.path.open("rb")
        self.zipfile = ZipFile(self.fp)
        return self

    def __exit__(self, _exc_type, _exc_value, _traceback):
        self.zipfile.close()
        self.fp.close()
        self.fp = None
        self.zipfile = None
        return False

    @property
    def dist_info(self):
        if self._dist_info is None:
            if self.zipfile is None:
                raise RuntimeError(
                    "WheelFile.dist_info cannot be determined when WheelFile"
                    " is not open in context"
                )
            self._dist_info = find_dist_info_dir(
                self.zipfile.namelist(),
                self.parsed_filename.project,
                self.parsed_filename.version,
            )
        return self._dist_info

    def basic_metadata(self):
        namebits = self.parsed_filename
        about = {
            "filename": self.path.name,
            "project": namebits.project,
            "version": namebits.version,
            "buildver": namebits.build,
            "pyver": namebits.python_tags,
            "abi": namebits.abi_tags,
            "arch": namebits.platform_tags,
            "file": {
                "size": self.path.stat().st_size,
            },
        }
        self.fp.seek(0)
        about["file"]["digests"] = digest_file(self.fp, ["md5", "sha256"])
        return about

    def open_dist_info_file(self, path):
        # returns a binary IO handle; raises MissingDistInfoFileError if file
        # does not exist
        try:
            zi = self.zipfile.getinfo(self.dist_info + "/" + path)
        except KeyError:
            raise errors.MissingDistInfoFileError(path)
        else:
            return self.zipfile.open(zi)

    def has_dist_info_file(self, path):  # -> bool
        try:
            self.zipfile.getinfo(self.dist_info + "/" + path)
        except KeyError:
            return False
        else:
            return True

    def list_files(self):
        return [name for name in self.zipfile.namelist() if not name.endswith("/")]

    def has_directory(self, path):
        return any(name.startswith(path) for name in self.zipfile.namelist())

    def get_file_size(self, path):
        return self.zipfile.getinfo(path).file_size

    def get_file_hash(self, path, algorithm):
        with self.zipfile.open(path) as fp:
            return digest_file(fp, [algorithm])[algorithm]



class WheelValidationError(Exception):
    """Superclass for all wheel validation errors raised by this package"""

    pass


class RecordValidationError(WheelValidationError):
    """
    Superclass for all validation errors raised due to a wheel's :file:`RECORD`
    being inaccurate or incomplete
    """

    pass


class RecordSizeMismatchError(RecordValidationError):
    """
    Raised when the size of a file as declared in a wheel's :file:`RECORD` does
    not match the file's actual size
    """

    def __init__(self, path, record_size, actual_size):
        #: The path of the mismatched file
        self.path = path
        #: The size of the file as declared in the :file:`RECORD`
        self.record_size = record_size
        #: The file's actual size
        self.actual_size = actual_size
        super().__init__(path, record_size, actual_size)

    def __str__(self):
        return (
            f"Size of file {self.path!r} listed as {self.record_size} in"
            f" RECORD, actually {self.actual_size}"
        )


class RecordDigestMismatchError(RecordValidationError):
    """
    Raised when a file's digest as declared in a wheel's :file:`RECORD` does
    not match the file's actual digest
    """

    def __init__(self, path, algorithm, record_digest, actual_digest):
        #: The path of the mismatched file
        self.path = path
        #: The name of the digest algorithm
        self.algorithm = algorithm
        #: The file's digest as declared in the :file:`RECORD`, in hex
        self.record_digest = record_digest
        #: The file's actual digest, in hex
        self.actual_digest = actual_digest
        super().__init__(path, algorithm, record_digest, actual_digest)

    def __str__(self):
        return (
            f"{self.algorithm} digest of file {self.path!r} listed as"
            f" {self.record_digest} in RECORD, actually {self.actual_digest}"
        )


class FileMissingError(RecordValidationError):
    """
    Raised when a file listed in a wheel's :file:`RECORD` is not found in the
    wheel
    """

    def __init__(self, path):
        #: The path of the missing file
        self.path = path
        super().__init__(path)

    def __str__(self):
        return "File declared in RECORD not found in archive: " + repr(self.path)


class ExtraFileError(RecordValidationError):
    """
    Raised when a wheel contains a file that is not listed in the
    :file:`RECORD` (other than :file:`RECORD.jws` and :file:`RECORD.p7s`)
    """

    def __init__(self, path):
        #: The path of the extra file
        self.path = path
        super().__init__(path)

    def __str__(self):
        return f"File not declared in RECORD: {self.path!r}"


class MalformedRecordError(WheelValidationError):
    """
    Superclass for all validation errors raised due to a wheel's :file:`RECORD`
    being malformed
    """

    pass


class UnknownDigestError(MalformedRecordError):
    """
    Raised when an entry in a wheel's :file:`RECORD` uses a digest not listed
    in `hashlib.algorithms_guaranteed`
    """

    def __init__(self, path, algorithm):
        #: The path the entry is for
        self.path = path
        #: The unknown digest algorithm
        self.algorithm = algorithm
        super().__init__(path, algorithm)

    def __str__(self):
        return (
            f"RECORD entry for {self.path!r} uses an unknown digest algorithm:"
            f" {self.algorithm!r}"
        )


class WeakDigestError(MalformedRecordError):
    """
    Raised when an entry in a wheel's :file:`RECORD` uses a digest weaker than
    sha256
    """

    def __init__(self, path, algorithm):
        #: The path the entry is for
        self.path = path
        #: The weak digest algorithm
        self.algorithm = algorithm
        super().__init__(path, algorithm)

    def __str__(self):
        return (
            f"RECORD entry for {self.path!r} uses a weak digest algorithm:"
            f" {self.algorithm!r}"
        )


class MalformedDigestError(MalformedRecordError):
    """
    Raised when an entry in a wheel's :file:`RECORD` contains a malformed or
    invalid digest
    """

    def __init__(self, path, algorithm, digest):
        #: The path the entry is for
        self.path = path
        #: The digest's declared algorithm
        self.algorithm = algorithm
        #: The malformed digest
        self.digest = digest
        super().__init__(path, algorithm, digest)

    def __str__(self):
        return (
            f"RECORD contains invalid {self.algorithm} base64 nopad digest for"
            f" {self.path!r}: {self.digest!r}"
        )


class MalformedSizeError(MalformedRecordError):
    """
    Raised when an entry in a wheel's :file:`RECORD` contains a malformed or
    invalid file size
    """

    def __init__(self, path, size):
        #: The path the entry is for
        self.path = path
        #: The size (as a string)
        self.size = size
        super().__init__(path, size)

    def __str__(self):
        return f"RECORD contains invalid size for {self.path!r}: {self.size!r}"


class RecordConflictError(MalformedRecordError):
    """
    Raised when a wheel's :file:`RECORD` contains two or more conflicting
    entries for the same path
    """

    def __init__(self, path):
        #: The path with conflicting entries
        self.path = path
        super().__init__(path)

    def __str__(self):
        return f"RECORD contains multiple conflicting entries for {self.path!r}"


class EmptyDigestError(MalformedRecordError):
    """
    Raised when an entry in a wheel's :file:`RECORD` has a size but not a
    digest
    """

    def __init__(self, path):
        #: The path the entry is for
        self.path = path
        super().__init__(path)

    def __str__(self):
        return f"RECORD entry for {self.path!r} has a size but no digest"


class EmptySizeError(MalformedRecordError):
    """
    Raised when an entry in a wheel's :file:`RECORD` has a digest but not a
    size
    """

    def __init__(self, path):
        #: The path the entry is for
        self.path = path
        super().__init__(path)

    def __str__(self):
        return f"RECORD entry for {self.path!r} has a digest but no size"


class EmptyPathError(MalformedRecordError):
    """Raised when an entry in a wheel's :file:`RECORD` has an empty path"""

    def __str__(self):
        return "RECORD entry has an empty path"


class RecordLengthError(MalformedRecordError):
    """
    Raised when an entry in a wheel's :file:`RECORD` has the wrong number of
    fields
    """

    def __init__(self, path, length):
        #: The path the entry is for (if nonempty)
        self.path = path
        #: The number of fields in the entry
        self.length = length
        super().__init__(path, length)

    def __str__(self):
        if self.path is None:
            return "Empty RECORD entry (blank line)"
        else:
            return (
                f"RECORD entry for {self.path!r} has {self.length} fields;"
                " expected 3"
            )


class NullEntryError(MalformedRecordError):
    """
    Raised when an entry in a wheel's :file:`RECORD` lacks both digest and size
    and the entry is not for the :file:`RECORD` itself
    """

    def __init__(self, path):
        #: The path the entry is for
        self.path = path
        super().__init__(path)

    def __str__(self):
        return f"RECORD entry for {self.path!r} lacks both digest and size"


class NonNormalizedPathError(MalformedRecordError):
    """
    Raised when an entry in a wheel's :file:`RECORD` has a non-normalized path
    """

    def __init__(self, path):
        #: The non-normalized path
        self.path = path
        super().__init__(path)

    def __str__(self):
        return f"RECORD entry has a non-normalized path: {self.path!r}"


class AbsolutePathError(MalformedRecordError):
    """
    Raised when an entry in a wheel's :file:`RECORD` has an absolute path
    """

    def __init__(self, path):
        #: The absolute path
        self.path = path
        super().__init__(path)

    def __str__(self):
        return f"RECORD entry has an absolute path: {self.path!r}"


class DistInfoError(WheelValidationError):
    """
    Raised when a wheel's :file:`*.dist-info` directory cannot be found or
    determined
    """

    pass


class MissingDistInfoFileError(WheelValidationError):
    """
    Raised when a given file is not found in the wheel's :file:`*.dist-info`
    directory
    """

    def __init__(self, path):
        #: The path to the file, relative to the :file:`*.dist-info` directory
        self.path = path
        super().__init__(path)

    def __str__(self):
        return f"File not found in *.dist-info directory: {self.path!r}"


class MissingMetadataError(MissingDistInfoFileError):
    """Raised when a wheel does not contain a :file:`METADATA` file"""

    def __init__(self):
        super().__init__("METADATA")


class MissingRecordError(MissingDistInfoFileError):
    """Raised when a wheel does not contain a :file:`RECORD` file"""

    def __init__(self):
        super().__init__("RECORD")


class MissingWheelInfoError(MissingDistInfoFileError):
    """Raised when a wheel does not contain a :file:`WHEEL` file"""

    def __init__(self):
        super().__init__("WHEEL")



from io import StringIO
import json
from operator import attrgetter
from pathlib import Path
import pytest
from wheel_inspect.errors import MalformedRecordError
from wheel_inspect.record import parse_record


def test_parse_record():
    assert (
        parse_record(
            StringIO(
                """\
qypi/__init__.py,sha256=zgE5-Sk8hED4NRmtnPUuvp1FDC4Z6VWCzJOOZwZ2oh8,532
qypi/__main__.py,sha256=GV5UVn3j5z4x-r7YYEB-quNPCucZYK1JOfWxmbdB0N0,7915
qypi/api.py,sha256=2c4EwxDhhHEloeOIeN0YgpIxCGpZaTDNJMYtHlVCcl8,3867
qypi/util.py,sha256=I2mRemqS5PHe5Iabk-CLrgFB2rznR87dVI3YwvpctSQ,3282

qypi-0.4.1.dist-info/DESCRIPTION.rst,sha256=SbT27FgdGvU8QlauLamstt7g4v7Cr2j6jc4RPr7bKNU,11633
qypi-0.4.1.dist-info/LICENSE.txt,sha256=SDaeT4Cm3ZeLgPOOL_f9BliMMHH_GVwqJa6czCztoS0,1090
qypi-0.4.1.dist-info/METADATA,sha256=msK-_0Fe8JHBjBv4HH35wbpUbIlCYv1Vy3X37tIdY5I,12633
qypi-0.4.1.dist-info/RECORD,,
qypi-0.4.1.dist-info/WHEEL,sha256=rNo05PbNqwnXiIHFsYm0m22u4Zm6YJtugFG2THx4w3g,92
qypi-0.4.1.dist-info/entry_points.txt,sha256=t4_O2VB3V-o52_PLoLLIb8m4SQDmY0HFdEJ9_Q2Odtw,45
qypi-0.4.1.dist-info/metadata.json,sha256=KI5TdfaYL-TPS1dMTABV6S8BFq9iAJRk3rkTXjOdgII,1297
qypi-0.4.1.dist-info/top_level.txt,sha256=J2Q5xVa8BtnOTGxjqY2lKQRB22Ydn9JF2PirqDEKE_Y,5
"""
            )
        ).for_json()
        == [
            {
                "path": "qypi/__init__.py",
                "digests": {"sha256": "zgE5-Sk8hED4NRmtnPUuvp1FDC4Z6VWCzJOOZwZ2oh8"},
                "size": 532,
            },
            {
                "path": "qypi/__main__.py",
                "digests": {"sha256": "GV5UVn3j5z4x-r7YYEB-quNPCucZYK1JOfWxmbdB0N0"},
                "size": 7915,
            },
            {
                "path": "qypi/api.py",
                "digests": {"sha256": "2c4EwxDhhHEloeOIeN0YgpIxCGpZaTDNJMYtHlVCcl8"},
                "size": 3867,
            },
            {
                "path": "qypi/util.py",
                "digests": {"sha256": "I2mRemqS5PHe5Iabk-CLrgFB2rznR87dVI3YwvpctSQ"},
                "size": 3282,
            },
            {
                "path": "qypi-0.4.1.dist-info/DESCRIPTION.rst",
                "digests": {"sha256": "SbT27FgdGvU8QlauLamstt7g4v7Cr2j6jc4RPr7bKNU"},
                "size": 11633,
            },
            {
                "path": "qypi-0.4.1.dist-info/LICENSE.txt",
                "digests": {"sha256": "SDaeT4Cm3ZeLgPOOL_f9BliMMHH_GVwqJa6czCztoS0"},
                "size": 1090,
            },
            {
                "path": "qypi-0.4.1.dist-info/METADATA",
                "digests": {"sha256": "msK-_0Fe8JHBjBv4HH35wbpUbIlCYv1Vy3X37tIdY5I"},
                "size": 12633,
            },
            {"path": "qypi-0.4.1.dist-info/RECORD", "digests": {}, "size": None},
            {
                "path": "qypi-0.4.1.dist-info/WHEEL",
                "digests": {"sha256": "rNo05PbNqwnXiIHFsYm0m22u4Zm6YJtugFG2THx4w3g"},
                "size": 92,
            },
            {
                "path": "qypi-0.4.1.dist-info/entry_points.txt",
                "digests": {"sha256": "t4_O2VB3V-o52_PLoLLIb8m4SQDmY0HFdEJ9_Q2Odtw"},
                "size": 45,
            },
            {
                "path": "qypi-0.4.1.dist-info/metadata.json",
                "digests": {"sha256": "KI5TdfaYL-TPS1dMTABV6S8BFq9iAJRk3rkTXjOdgII"},
                "size": 1297,
            },
            {
                "path": "qypi-0.4.1.dist-info/top_level.txt",
                "digests": {"sha256": "J2Q5xVa8BtnOTGxjqY2lKQRB22Ydn9JF2PirqDEKE_Y"},
                "size": 5,
            },
        ]
    )


@pytest.mark.parametrize(
    "recfile",
    (Path(__file__).with_name("data") / "bad-records").glob("*.csv"),
    ids=attrgetter("name"),
)
def test_parse_bad_records(recfile):
    with recfile.with_suffix(".json").open() as fp:
        expected = json.load(fp)
    with recfile.open(newline="") as fp:
        with pytest.raises(MalformedRecordError) as excinfo:
            parse_record(fp)
        assert type(excinfo.value).__name__ == expected["type"]
        assert str(excinfo.value) == expected["str"]



from copy import deepcopy

#: A `JSON Schema <http://json-schema.org>`_ for the structure returned by
#: `inspect_dist_info_dir()` and by `inspect()` when called on a `DistInfoDir`.
#: It is the same as `WHEEL_SCHEMA`, but without the ``"filename"``,
#: ``"project"``, ``"version"``, ``"buildver"``, ``"pyver"``, ``"abi"``,
#: ``"arch"``, and ``"file"`` keys.
DIST_INFO_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "type": "object",
    "required": ["valid", "dist_info", "derived"],
    "additionalProperties": False,
    "properties": {
        "valid": {
            "type": "boolean",
            "description": "Whether the wheel is well-formed with an accurate RECORD",
        },
        "validation_error": {
            "type": "object",
            "description": "If the wheel is invalid, this field contains information on the `WheelValidationError` raised.",
            "required": ["type", "str"],
            "additionalProperties": False,
            "properties": {
                "type": {
                    "type": "string",
                    "description": "The name of the type of exception raised",
                },
                "str": {
                    "type": "string",
                    "description": "The exception's error message",
                },
            },
        },
        "dist_info": {
            "type": "object",
            "description": "JSONifications of files from the wheel's .dist-info directory",
            "additionalProperties": False,
            "properties": {
                "metadata": {
                    "type": "object",
                    "additionalProperties": {
                        "type": "array",
                        "items": {"type": "string"},
                    },
                    "properties": {
                        "metadata_version": {"type": "string"},
                        "name": {"type": "string"},
                        "version": {"type": "string"},
                        "summary": {"type": ["null", "string"]},
                        "description": {
                            "oneOf": [
                                {"type": "null"},
                                {
                                    "type": "object",
                                    "requires": ["length"],
                                    "additionalProperties": False,
                                    "properties": {"length": {"type": "integer"}},
                                },
                            ]
                        },
                        "requires_dist": {
                            "type": "array",
                            "items": {
                                "type": "object",
                                "required": [
                                    "name",
                                    "url",
                                    "extras",
                                    "specifier",
                                    "marker",
                                ],
                                "additionalProperties": False,
                                "properties": {
                                    "name": {"type": "string"},
                                    "url": {"type": ["null", "string"]},
                                    "extras": {
                                        "type": "array",
                                        "items": {"type": "string"},
                                    },
                                    "specifier": {"type": "string"},
                                    "marker": {"type": ["null", "string"]},
                                },
                            },
                        },
                        "project_url": {
                            "type": "array",
                            "items": {
                                "type": "object",
                                "requires": ["label", "url"],
                                "additionalProperties": False,
                                "properties": {
                                    "label": {"type": ["null", "string"]},
                                    "url": {"type": "string"},
                                },
                            },
                        },
                        "requires_python": {"type": "string"},
                        "author": {"type": ["null", "string"]},
                        "author_email": {"type": ["null", "string"]},
                        "download_url": {"type": ["null", "string"]},
                        "home_page": {"type": ["null", "string"]},
                        "license": {"type": ["null", "string"]},
                        "maintainer": {"type": ["null", "string"]},
                        "maintainer_email": {"type": ["null", "string"]},
                        "keywords": {"type": ["null", "string"]},
                        "description_content_type": {"type": ["null", "string"]},
                    },
                },
                "record": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "required": ["path", "digests", "size"],
                        "additionalProperties": False,
                        "properties": {
                            "path": {"type": "string"},
                            "digests": {
                                "type": "object",
                                "properties": {
                                    "md5": {
                                        "type": "string",
                                        "pattern": "^[-_0-9A-Za-z]{22}$",
                                    },
                                    "sha1": {
                                        "type": "string",
                                        "pattern": "^[-_0-9A-Za-z]{27}$",
                                    },
                                    "sha256": {
                                        "type": "string",
                                        "pattern": "^[-_0-9A-Za-z]{43}$",
                                    },
                                    "sha512": {
                                        "type": "string",
                                        "pattern": "^[-_0-9A-Za-z]{86}$",
                                    },
                                },
                            },
                            "size": {"type": ["null", "integer"]},
                        },
                    },
                },
                "wheel": {
                    "type": "object",
                    "required": [
                        "wheel_version",
                        "generator",
                        "root_is_purelib",
                        "tag",
                    ],
                    "additionalProperties": {
                        "type": "array",
                        "items": {"type": "string"},
                    },
                    "properties": {
                        "wheel_version": {"type": "string"},
                        "generator": {"type": "string"},
                        "root_is_purelib": {"type": "boolean"},
                        "tag": {"type": "array", "items": {"type": "string"}},
                        "build": {"type": "string"},
                        "BODY": {"type": "string"},
                    },
                },
                "dependency_links": {"type": "array", "items": {"type": "string"}},
                "entry_points": {
                    "type": "object",
                    "additionalProperties": {
                        "type": "object",
                        "additionalProperties": {
                            "type": "object",
                            "required": ["module", "attr", "extras"],
                            "additionalProperties": False,
                            "properties": {
                                "module": {"type": "string"},
                                "attr": {"type": ["null", "string"]},
                                "extras": {
                                    "type": "array",
                                    "items": {"type": "string"},
                                },
                            },
                        },
                    },
                },
                "namespace_packages": {"type": "array", "items": {"type": "string"}},
                "top_level": {"type": "array", "items": {"type": "string"}},
                "zip_safe": {"type": "boolean"},
            },
        },
        "derived": {
            "type": "object",
            "description": "Information derived from `dist_info`",
            "required": [
                "readme_renders",
                "description_in_body",
                "description_in_headers",
                "keywords",
                "keyword_separator",
                "dependencies",
                "modules",
            ],
            "additionalProperties": False,
            "properties": {
                "readme_renders": {
                    "type": ["null", "boolean"],
                    "description": "Whether the description's markup can be rendered successfully on PyPI.  A value of `null` indicates that there is no description.",
                },
                "description_in_body": {
                    "type": "boolean",
                    "description": "Whether the description is present as the message body in the METADATA file",
                },
                "description_in_headers": {
                    "type": "boolean",
                    "description": "Whether the description is present as a header field in the METADATA file",
                },
                "keywords": {
                    "type": "array",
                    "items": {"type": "string"},
                    "uniqueItems": True,
                    "description": "The wheel's keywords string, split on what appears to be the appropriate separator",
                },
                "keyword_separator": {
                    "enum": [" ", ",", None],
                    "description": "The apparent appropriate separator for the wheel's keywords string.  A value of `null` indicates that the keywords string is undefined.",
                },
                "dependencies": {
                    "type": "array",
                    "items": {"type": "string"},
                    "uniqueItems": True,
                    "description": "The names of all of the projects listed in the wheel's Requires-Dist",
                },
                "modules": {
                    "type": "array",
                    "items": {"type": "string"},
                    "uniqueItems": True,
                    "description": "A list of Python modules installed by the wheel",
                },
            },
        },
    },
}

#: A `JSON Schema <http://json-schema.org>`_ for the structure returned by
#: `inspect_wheel()` and by `inspect()` when called on a `WheelFile`.
WHEEL_SCHEMA = deepcopy(DIST_INFO_SCHEMA)

WHEEL_SCHEMA["required"].extend(
    [
        "filename",
        "project",
        "version",
        "buildver",
        "pyver",
        "abi",
        "arch",
        "file",
    ]
)

WHEEL_SCHEMA["properties"].update(
    {
        "filename": {"type": "string", "description": "The filename of the wheel"},
        "project": {
            "type": "string",
            "description": "The name of the wheel's project as extracted from the filename",
        },
        "version": {
            "type": "string",
            "description": "The wheel's project version as extracted from the filename",
        },
        "buildver": {
            "type": ["string", "null"],
            "description": "The wheel's build tag as extracted from the filename; `null` if there is no build tag",
        },
        "pyver": {
            "type": "array",
            "items": {"type": "string"},
            "description": "A list of Python versions with which the wheel is compatible as extracted from the filename",
        },
        "abi": {
            "type": "array",
            "items": {"type": "string"},
            "description": "A list of ABIs with which the wheel is compatible as extracted from the filename",
        },
        "arch": {
            "type": "array",
            "items": {"type": "string"},
            "description": "A list of architectures with which the wheel is compatible as extracted from the filename",
        },
        "file": {
            "type": "object",
            "required": ["size", "digests"],
            "additionalProperties": False,
            "properties": {
                "size": {"type": "integer"},
                "digests": {
                    "type": "object",
                    "required": ["md5", "sha256"],
                    "additionalProperties": False,
                    "properties": {
                        "md5": {"type": "string", "pattern": "^[0-9A-Fa-f]{32}$"},
                        "sha256": {"type": "string", "pattern": "^[0-9A-Fa-f]{64}$"},
                    },
                },
            },
        },
    }
)

#: Alias for `WHEEL_SCHEMA`.  Deprecated; use `WHEEL_SCHEMA` instead.
SCHEMA = WHEEL_SCHEMA



from headerparser import BOOL, HeaderParser
from .util import fieldnorm, strfield

infoparser = HeaderParser(normalizer=fieldnorm)
# NOTE: WHEEL files generated by `wheel` (but not by `flit`) end with a blank
# line, which headerparser interprets as the start of an empty body.
infoparser.add_field("Wheel-Version", required=True)
infoparser.add_field("Generator", required=True)
infoparser.add_field("Root-Is-Purelib", required=True, type=BOOL)
infoparser.add_field("Tag", required=True, multiple=True)
infoparser.add_field("Build")
infoparser.add_additional(multiple=True, type=strfield)


def parse_wheel_info(fp):
    """
    Parse the contents of a text filehandle ``fp`` as a
    :file:`*.dist-info/WHEEL` file and return a `dict` that maps field names to
    field name values.  All field names are converted to lowercase and have
    hyphens changed to underscores.  Field values are strings, except for
    :mailheader:`Root-Is-Purelib`, whose value is converted to a `bool`, and
    except for :mailheader:`Tag` and any unknown fields, whose values are lists
    of strings.  (Known field names are :mailheader:`Wheel-Version`,
    :mailheader:`Generator`, :mailheader:`Root-Is-Purelib`, :mailheader:`Tag`,
    and :mailheader:`Build`; all but :mailheader:`Build` are required.)  If the
    input contains a message body that is not all whitespace, it is stored
    under the ``"BODY"`` key (which is otherwise absent).

    For example, the following input:

    .. code-block:: email

        Wheel-Version: 1.0
        Generator: bdist_wheel (0.29.0)
        Root-Is-Purelib: true
        Tag: py3-none-any
        Tag: py2-none-any
        Unknown-Field: foo

        This is a body.

    would be parsed into the following structure::

        {
            "wheel_version": "1.0",
            "generator": "bdist_wheel (0.29.0)",
            "root_is_purelib": True,
            "tag": ["py3-none-any", "py2-none-any"],
            "unknown_field": ["foo"],
            "BODY": "This is a body.\\n"
        }
    """
    wi = infoparser.parse(fp)
    wheel_info = wi.normalized_dict()
    for k, v in wheel_info.items():
        if isinstance(v, list):
            wheel_info[k] = [u for u in v if u is not None]
    if wi.body is not None and wi.body.strip():
        wheel_info["BODY"] = wi.body
    return wheel_info



from email.message import EmailMessage
import hashlib
from keyword import iskeyword
import re
from packaging.utils import canonicalize_name, canonicalize_version
from .errors import DistInfoError

DIGEST_CHUNK_SIZE = 65535

DIST_INFO_DIR_RGX = re.compile(
    r"[A-Za-z0-9](?:[A-Za-z0-9._]*[A-Za-z0-9])?-[A-Za-z0-9_.!+]+\.dist-info"
)

DATA_DIR_RGX = re.compile(
    r"[A-Za-z0-9](?:[A-Za-z0-9._]*[A-Za-z0-9])?-[A-Za-z0-9_.!+]+\.data"
)

# <https://discuss.python.org/t/identifying-parsing-binary-extension-filenames/>
MODULE_EXT_RGX = re.compile(r"(?<=.)\.(?:py|pyd|so|[-A-Za-z0-9_]+\.(?:pyd|so))\Z")


def extract_modules(filelist):
    modules = set()
    for fname in filelist:
        parts = fname.split("/")
        if not parts:
            continue
        if (
            len(parts) > 2
            and is_data_dir(parts[0])
            and parts[1] in ("purelib", "platlib")
        ):
            parts = parts[2:]
        m = MODULE_EXT_RGX.search(parts[-1])
        if m is None:
            continue
        parts[-1] = parts[-1][: m.start()]
        if not all(p.isidentifier() and not iskeyword(p) for p in parts):
            continue
        if parts[-1] == "__init__" and len(parts) > 1:
            parts.pop()
        modules.add(".".join(parts))
    return sorted(modules)


def split_keywords(kwstr):
    # cf. `format_tags()` in Warehouse <https://git.io/fA1AT>, which seems to
    # be the part of PyPI responsible for splitting keywords up for display

    # cf. how wheel handles keywords:
    # keywords = re.split(r'[\0-,]+', kwstr)

    # Based on how pydigger.com seems to handle keywords (See
    # <https://pydigger.com/keywords>):
    if "," in kwstr:
        return ([k for k in map(str.strip, kwstr.split(",")) if k], ",")
    else:
        return (kwstr.split(), " ")


def strfield(s):
    return None if s is None or s.strip() in ("", "UNKNOWN") else s


def fieldnorm(s):
    return s.lower().replace("-", "_")


def unique_projects(projects):
    seen = set()
    for p in projects:
        pn = canonicalize_name(p)
        if pn not in seen:
            yield p
        seen.add(pn)


def digest_file(fp, algorithms):
    digests = {alg: getattr(hashlib, alg)() for alg in algorithms}
    for chunk in iter(lambda: fp.read(DIGEST_CHUNK_SIZE), b""):
        for d in digests.values():
            d.update(chunk)
    return {k: v.hexdigest() for k, v in digests.items()}


def split_content_type(s):
    msg = EmailMessage()
    msg["Content-Type"] = s
    ct = msg["Content-Type"]
    return (ct.maintype, ct.subtype, ct.params)


def is_dist_info_dir(name):
    return DIST_INFO_DIR_RGX.fullmatch(name) is not None


def is_data_dir(name):
    return DATA_DIR_RGX.fullmatch(name) is not None


def is_dist_info_path(path, name):
    pre, _, post = path.partition("/")
    return is_dist_info_dir(pre) and post == name


def yield_lines(fp):
    # Like pkg_resources.yield_lines(fp), but without the dependency on
    # pkg_resources
    for line in fp:
        line = line.strip()
        if line and not line.startswith("#"):
            yield line


def find_dist_info_dir(
    namelist: list[str], project: str, version: str
) -> tuple[str, str | None]:
    """
    Given a list ``namelist`` of files in a wheel for a project ``project`` and
    version ``version``, find & return the name of the wheel's ``.dist-info``
    directory.

    :raises DistInfoError: if there is no unique ``.dist-info`` directory in
        the input
    :raises DistInfoError: if the name & version of the ``.dist-info``
        directory are not normalization-equivalent to ``project`` & ``version``
    """
    canon_project = canonicalize_name(project)
    canon_version = canonicalize_version(version.replace("_", "-"))
    dist_info_dirs = set()
    for n in namelist:
        basename = n.rstrip("/").split("/")[0]
        if is_dist_info_dir(basename):
            dist_info_dirs.add(basename)
    if len(dist_info_dirs) > 1:
        raise DistInfoError("Wheel contains multiple .dist-info directories")
    elif len(dist_info_dirs) == 1:
        dist_info_dir = next(iter(dist_info_dirs))
        diname, _, diversion = dist_info_dir[: -len(".dist-info")].partition("-")
        if (
            canonicalize_name(diname) != canon_project
            or canonicalize_version(diversion.replace("_", "-")) != canon_version
        ):
            raise DistInfoError(
                f"Project & version of wheel's .dist-info directory do not"
                f" match wheel name: {dist_info_dir!r}"
            )
        return dist_info_dir
    else:
        raise DistInfoError("No .dist-info directory in wheel")



# cf. PEP 345 and <https://packaging.python.org/specifications/>
import re
from headerparser import HeaderParser
from packaging.requirements import Requirement
from .util import fieldnorm, strfield


def requirement(s):
    req = Requirement(s)
    return {
        "name": req.name,
        "url": req.url,
        "extras": sorted(req.extras),
        "specifier": str(req.specifier),
        "marker": str(req.marker) if req.marker is not None else None,
    }


def project_url(s):
    try:
        label, url = re.split(r"\s*,\s*", s, maxsplit=1)
    except ValueError:
        label, url = None, s
    return {"label": label, "url": url}


metaparser = HeaderParser(normalizer=fieldnorm)
metaparser.add_field("Metadata-Version")
metaparser.add_field("Name")
metaparser.add_field("Version")
metaparser.add_field("Summary", type=strfield)
metaparser.add_field("Requires-Dist", type=requirement, multiple=True)
metaparser.add_field("Requires-Python")
metaparser.add_field("Project-URL", type=project_url, multiple=True)

for field in (
    "Author Author-email Description Download-URL Home-page License"
    " Maintainer Maintainer-email Keywords Description-Content-Type".split()
):
    ### TODO: Dedent Description (et alii?)?
    metaparser.add_field(field, type=strfield)

for field in (
    "Classifier Obsoletes Obsoletes-Dist Platform Provides"
    " Provides-Dist Provides-Extra Requires Requires-External"
    " Supported-Platform".split()
):
    metaparser.add_field(field, type=strfield, multiple=True)

metaparser.add_additional(multiple=True, type=strfield)


def parse_metadata(fp):
    md = metaparser.parse(fp)
    metadata = md.normalized_dict()
    for k, v in metadata.items():
        if isinstance(v, list):
            metadata[k] = [u for u in v if u is not None]
    if md.body is not None:
        metadata["BODY"] = strfield(md.body)
    return metadata



"""
Extract information from wheels

``wheel-inspect`` examines Python wheel files and outputs various information
about the contents within as JSON-serializable objects.  It can be invoked in
Python code as::

    from wheel_inspect import inspect_wheel

    output = inspect_wheel(path_to_wheel_file)

or from the command line with the ``wheel2json`` command.

Visit <https://github.com/wheelodex/wheel-inspect> for more information.
"""

from .inspecting import inspect_dist_info_dir, inspect_wheel
from .schema import DIST_INFO_SCHEMA, SCHEMA, WHEEL_SCHEMA

__version__ = "1.8.0"
__author__ = "John Thorvald Wodder II"
__author_email__ = "wheel-inspect@varonathe.org"
__license__ = "MIT"
__url__ = "https://github.com/wheelodex/wheel-inspect"

__all__ = [
    "DIST_INFO_SCHEMA",
    "SCHEMA",
    "WHEEL_SCHEMA",
    "inspect_dist_info_dir",
    "inspect_wheel",
]



import pytest
from wheel_inspect.errors import DistInfoError
from wheel_inspect.util import (
    extract_modules,
    find_dist_info_dir,
    is_data_dir,
    is_dist_info_dir,
    split_content_type,
    split_keywords,
)


@pytest.mark.parametrize(
    "kwstr,expected",
    [
        (
            "pypi,warehouse,search,packages,pip",
            (["pypi", "warehouse", "search", "packages", "pip"], ","),
        ),
        (
            "pypi warehouse search packages pip",
            (["pypi", "warehouse", "search", "packages", "pip"], " "),
        ),
        (
            "pypi,pep503,simple repository api,packages,pip",
            (["pypi", "pep503", "simple repository api", "packages", "pip"], ","),
        ),
        ("", ([], " ")),
        (" ", ([], " ")),
        (",", ([], ",")),
        (" , ", ([], ",")),
        (" , , ", ([], ",")),
        ("foo", (["foo"], " ")),
        ("foo,bar", (["foo", "bar"], ",")),
        ("foo, bar", (["foo", "bar"], ",")),
        ("foo ,bar", (["foo", "bar"], ",")),
        (" foo , bar ", (["foo", "bar"], ",")),
        (" foo , , bar ", (["foo", "bar"], ",")),
        ("foo,,bar", (["foo", "bar"], ",")),
        (",foo", (["foo"], ",")),
        ("foo,", (["foo"], ",")),
    ],
)
def test_split_keywords(kwstr, expected):
    assert split_keywords(kwstr) == expected


@pytest.mark.parametrize(
    "filelist,modules",
    [
        (
            [
                "qypi/__init__.py",
                "qypi/__main__.py",
                "qypi/api.py",
                "qypi/util.py",
                "qypi-0.4.1.dist-info/DESCRIPTION.rst",
                "qypi-0.4.1.dist-info/LICENSE.txt",
                "qypi-0.4.1.dist-info/METADATA",
                "qypi-0.4.1.dist-info/RECORD",
                "qypi-0.4.1.dist-info/WHEEL",
                "qypi-0.4.1.dist-info/entry_points.txt",
                "qypi-0.4.1.dist-info/metadata.json",
                "qypi-0.4.1.dist-info/top_level.txt",
            ],
            ["qypi", "qypi.__main__", "qypi.api", "qypi.util"],
        ),
        (
            [
                "flit/__init__.py",
                "flit/__main__.py",
                "flit/_get_dirs.py",
                "flit/build.py",
                "flit/common.py",
                "flit/inifile.py",
                "flit/init.py",
                "flit/install.py",
                "flit/installfrom.py",
                "flit/log.py",
                "flit/logo.py",
                "flit/sdist.py",
                "flit/upload.py",
                "flit/wheel.py",
                "flit/license_templates/apache",
                "flit/license_templates/gpl3",
                "flit/license_templates/mit",
                "flit/vcs/__init__.py",
                "flit/vcs/git.py",
                "flit/vcs/hg.py",
                "flit/vendorized/__init__.py",
                "flit/vendorized/readme/__init__.py",
                "flit/vendorized/readme/clean.py",
                "flit/vendorized/readme/rst.py",
                "flit-0.11.1.dist-info/entry_points.txt",
                "flit-0.11.1.dist-info/LICENSE",
                "flit-0.11.1.dist-info/WHEEL",
                "flit-0.11.1.dist-info/METADATA",
                "flit-0.11.1.dist-info/RECORD",
            ],
            [
                "flit",
                "flit.__main__",
                "flit._get_dirs",
                "flit.build",
                "flit.common",
                "flit.inifile",
                "flit.init",
                "flit.install",
                "flit.installfrom",
                "flit.log",
                "flit.logo",
                "flit.sdist",
                "flit.upload",
                "flit.vcs",
                "flit.vcs.git",
                "flit.vcs.hg",
                "flit.vendorized",
                "flit.vendorized.readme",
                "flit.vendorized.readme.clean",
                "flit.vendorized.readme.rst",
                "flit.wheel",
            ],
        ),
        (
            [
                "cmarkgfm-0.4.2.dist-info/WHEEL",
                "cmarkgfm-0.4.2.dist-info/top_level.txt",
                "cmarkgfm-0.4.2.dist-info/METADATA",
                "cmarkgfm-0.4.2.dist-info/RECORD",
                "cmarkgfm/_cmark.abi3.so",
                "cmarkgfm/cmark_module.h",
                "cmarkgfm/__init__.py",
                "cmarkgfm/cmark.py",
                "cmarkgfm/build_cmark.py",
                "cmarkgfm/cmark.cffi.h",
            ],
            [
                "cmarkgfm",
                "cmarkgfm._cmark",
                "cmarkgfm.build_cmark",
                "cmarkgfm.cmark",
            ],
        ),
        (
            [
                "cmarkgfm-0.4.2.dist-info/WHEEL",
                "cmarkgfm-0.4.2.dist-info/top_level.txt",
                "cmarkgfm-0.4.2.dist-info/METADATA",
                "cmarkgfm-0.4.2.dist-info/RECORD",
                "cmarkgfm/_cmark.cp37-win_amd64.pyd",
                "cmarkgfm/cmark_module.h",
                "cmarkgfm/__init__.py",
                "cmarkgfm/cmark.py",
                "cmarkgfm/build_cmark.py",
                "cmarkgfm/cmark.cffi.h",
            ],
            [
                "cmarkgfm",
                "cmarkgfm._cmark",
                "cmarkgfm.build_cmark",
                "cmarkgfm.cmark",
            ],
        ),
        (
            [
                "cmarkgfm-0.4.2.dist-info/WHEEL",
                "cmarkgfm-0.4.2.dist-info/top_level.txt",
                "cmarkgfm-0.4.2.dist-info/METADATA",
                "cmarkgfm-0.4.2.dist-info/RECORD",
                "cmarkgfm/_cmark.so",
                "cmarkgfm/cmark_module.h",
                "cmarkgfm/__init__.py",
                "cmarkgfm/cmark.py",
                "cmarkgfm/build_cmark.py",
                "cmarkgfm/cmark.cffi.h",
            ],
            [
                "cmarkgfm",
                "cmarkgfm._cmark",
                "cmarkgfm.build_cmark",
                "cmarkgfm.cmark",
            ],
        ),
        (
            [
                "mxnet_coreml_converter-0.1.0a7.data/purelib/converter/__init__.py",
                "mxnet_coreml_converter-0.1.0a7.data/purelib/converter/_add_pooling.py",
                "mxnet_coreml_converter-0.1.0a7.data/purelib/converter/_layers.py",
                "mxnet_coreml_converter-0.1.0a7.data/purelib/converter/_mxnet_converter.py",
                "mxnet_coreml_converter-0.1.0a7.data/purelib/converter/utils.py",
                "mxnet_coreml_converter-0.1.0a7.data/scripts/mxnet_coreml_converter.py",
                "mxnet_coreml_converter-0.1.0a7.dist-info/DESCRIPTION.rst",
                "mxnet_coreml_converter-0.1.0a7.dist-info/metadata.json",
                "mxnet_coreml_converter-0.1.0a7.dist-info/top_level.txt",
                "mxnet_coreml_converter-0.1.0a7.dist-info/WHEEL",
                "mxnet_coreml_converter-0.1.0a7.dist-info/METADATA",
                "mxnet_coreml_converter-0.1.0a7.dist-info/RECORD",
            ],
            [
                "converter",
                "converter._add_pooling",
                "converter._layers",
                "converter._mxnet_converter",
                "converter.utils",
            ],
        ),
        (
            [
                "",
                "foo-1.0.data/platlib/foo/__init__.py",
                "foo-1.0.data/platlib/foo/def.py",
                "foo-1.0.data/platlib/foo/has-hyphen.py",
                "foo-1.0.data/platlib/foo/extra.ext.py",
                "foo-1.0.dist-info/METADATA",
                "foo-1.0.dist-info/WHEEL",
                "foo-1.0.dist-info/RECORD",
                "foo-1.0.dist-info/glarch.py",
            ],
            ["foo"],
        ),
        (
            [
                "Acquisition/Acquisition.h",
                "Acquisition/_Acquisition.c",
                "Acquisition/_Acquisition.pyd",
                "Acquisition/__init__.py",
                "Acquisition/interfaces.py",
                "Acquisition/tests.py",
                "Acquisition-4.6.dist-info/LICENSE.txt",
                "Acquisition-4.6.dist-info/METADATA",
                "Acquisition-4.6.dist-info/WHEEL",
                "Acquisition-4.6.dist-info/top_level.txt",
                "Acquisition-4.6.dist-info/RECORD",
            ],
            [
                "Acquisition",
                "Acquisition._Acquisition",
                "Acquisition.interfaces",
                "Acquisition.tests",
            ],
        ),
    ],
)
def test_extract_modules(filelist, modules):
    assert extract_modules(filelist) == modules


@pytest.mark.parametrize(
    "s,ct",
    [
        ("text/plain", ("text", "plain", {})),
        ("text/plain; charset=utf-8", ("text", "plain", {"charset": "utf-8"})),
        (
            "text/markdown; charset=utf-8; variant=GFM",
            ("text", "markdown", {"charset": "utf-8", "variant": "GFM"}),
        ),
    ],
)
def test_split_content_type(s, ct):
    assert split_content_type(s) == ct


@pytest.mark.parametrize(
    "name,expected",
    [
        ("somepackage-1.0.0.dist-info", True),
        ("somepackage.dist-info", False),
        ("somepackage-1.0.0-1.dist-info", False),
        ("somepackage-1.0.0.data", False),
        ("SOME_._PaCkAgE-0.dist-info", True),
        ("foo-1!2+local.dist-info", True),
        ("foo-1_2_local.dist-info", True),
        (".dist-info", False),
    ],
)
def test_is_dist_info_dir(name, expected):
    assert is_dist_info_dir(name) is expected


@pytest.mark.parametrize(
    "name,expected",
    [
        ("somepackage-1.0.0.data", True),
        ("somepackage.data", False),
        ("somepackage-1.0.0-1.data", False),
        ("somepackage-1.0.0.dist-info", False),
        ("SOME_._PaCkAgE-0.data", True),
        ("foo-1!2+local.data", True),
        ("foo-1_2_local.data", True),
        (".data", False),
    ],
)
def test_is_data_dir(name, expected):
    assert is_data_dir(name) is expected


@pytest.mark.parametrize(
    "namelist,project,version,expected",
    [
        (
            [
                "foo.py",
                "foo-1.0.dist-info/WHEEL",
                "foo-1.0.dist-info/RECORD",
            ],
            "foo",
            "1.0",
            "foo-1.0.dist-info",
        ),
        (
            [
                "foo.py",
                "FOO-1.0.0.dist-info/WHEEL",
                "FOO-1.0.0.dist-info/RECORD",
            ],
            "foo",
            "1.0",
            "FOO-1.0.0.dist-info",
        ),
        (
            [
                "foo.py",
                "foo-1.dist-info/WHEEL",
                "foo-1.dist-info/RECORD",
            ],
            "foo",
            "1.0",
            "foo-1.dist-info",
        ),
        (
            [
                "foo.py",
                "FOO-1.0_1.dist-info/WHEEL",
                "FOO-1.0_1.dist-info/RECORD",
            ],
            "foo",
            "1.0.post1",
            "FOO-1.0_1.dist-info",
        ),
    ],
)
def test_find_dist_info_dir(namelist, project, version, expected):
    assert find_dist_info_dir(namelist, project, version) == expected


@pytest.mark.parametrize(
    "namelist,project,version,msg",
    [
        (
            [
                "foo.py",
                "foo-1.0.dist/WHEEL",
            ],
            "foo",
            "1.0",
            "No .dist-info directory in wheel",
        ),
        (
            [
                "foo.py",
                "bar-1.0.dist-info/WHEEL",
            ],
            "foo",
            "1.0",
            "Project & version of wheel's .dist-info directory do not match wheel"
            " name: 'bar-1.0.dist-info'",
        ),
        (
            [
                "foo.py",
                "foo-2.0.dist-info/WHEEL",
            ],
            "foo",
            "1.0",
            "Project & version of wheel's .dist-info directory do not match wheel"
            " name: 'foo-2.0.dist-info'",
        ),
        (
            [
                "foo.py",
                "foo-1.0.dist-info/WHEEL",
                "bar-2.0.dist-info/RECORD",
            ],
            "foo",
            "1.0",
            "Wheel contains multiple .dist-info directories",
        ),
        (
            [
                "foo.py",
                "FOO-1.0.0.dist-info/WHEEL",
                "foo-1.dist-info/RECORD",
            ],
            "foo",
            "1.0",
            "Wheel contains multiple .dist-info directories",
        ),
        (
            ["foo.py", ".dist-info/WHEEL"],
            "foo",
            "1.0",
            "No .dist-info directory in wheel",
        ),
    ],
)
def test_find_dist_info_dir_error(namelist, project, version, msg):
    with pytest.raises(DistInfoError) as excinfo:
        find_dist_info_dir(namelist, project, version)
    assert str(excinfo.value) == msg


### TODO: Add more test cases for all functions!



import json
import os.path
import sys
from .inspecting import inspect_dist_info_dir, inspect_wheel


def main():
    for path in sys.argv[1:]:
        if os.path.isdir(path):
            about = inspect_dist_info_dir(path)
        else:
            about = inspect_wheel(path)
        print(json.dumps(about, sort_keys=True, indent=4))


if __name__ == "__main__":
    main()



import json
from pathlib import Path
import pytest

DATA_DIR = Path(__file__).with_name("data")


def filecases(subdir, glob_pattern):
    for p in sorted((DATA_DIR / subdir).glob(glob_pattern)):
        with p.with_suffix(".json").open() as fp:
            expected = json.load(fp)
        yield pytest.param(p, expected, id=p.name)



import pytest
from testing_lib import filecases
from wheel_inspect.inspecting import parse_entry_points
from wheel_inspect.metadata import parse_metadata
from wheel_inspect.wheel_info import parse_wheel_info


@pytest.mark.parametrize("mdfile,expected", filecases("metadata", "*.metadata"))
def test_parse_metadata(mdfile, expected):
    with mdfile.open(encoding="utf-8") as fp:
        assert parse_metadata(fp) == expected


@pytest.mark.parametrize("epfile,expected", filecases("entry_points", "*.txt"))
def test_parse_entry_points(epfile, expected):
    with epfile.open(encoding="utf-8") as fp:
        assert parse_entry_points(fp) == expected


@pytest.mark.parametrize("wifile,expected", filecases("wheel_info", "*.wheel"))
def test_parse_wheel_info(wifile, expected):
    with wifile.open(encoding="utf-8") as fp:
        assert parse_wheel_info(fp) == expected
